// ( costanti usate in sistema.cpp e sistema.S
//   SEL = selettori (segmentazione con modello flat)
//   LIV = livelli di privilegio
#define SEL_CODICE_SISTEMA	0x8
#define SEL_CODICE_UTENTE	0x13
#define SEL_DATI_UTENTE 	0x1b
#define LIV_UTENTE		3
#define LIV_SISTEMA		0
#define NUM_TSS 		1024
#define MIN_PROC_ID		32
#define MAX_PROC_ID		(NUM_TSS*16)
// )

// ( varie dimensioni
#define KiB			1024UL
#define MiB			(1024*KiB)
#define GiB			(1024*MiB)

// dimensione in byte della parte hardware del descrittore di processo
#define DIM_DESP                104
#define MAX_SEM			4096
#define DIM_PAGINA		4096UL
#define MAX_PRD			16

#define MEM_TOT			(8*MiB)
#define DIM_USR_HEAP		(1*MiB)
#define DIM_USR_STACK		(10*MiB)
#define DIM_IO_HEAP		(1*MiB)
#define DIM_SYS_STACK		(4*KiB)
#define DIM_SWAP		(80*MiB)
#define DIM_BLOCK		512
// )

// ( tipi interruzioni esterne
#define VETT_0			0xF0
#define VETT_1                  0xD0
#define VETT_2                  0xC0
#define VETT_3                  0xB0
#define VETT_4                  0xA0
#define VETT_5                  0x90
#define VETT_6                  0x80
#define VETT_7                  0x70
#define VETT_8                  0x60
#define VETT_9                  0x50
#define VETT_10                 0x40
#define VETT_11                 0x30
#define VETT_12                 0x20
#define VETT_13                 0xD1
#define VETT_14                 0xE0
#define VETT_15                 0xE1
#define VETT_16                 0xC1
#define VETT_17                 0xB1
#define VETT_18                 0xA1
#define VETT_19                 0x91
#define VETT_20                 0x81
#define VETT_21                 0x71
#define VETT_22                 0x61
#define VETT_23                 0x51
#define VETT_S			0x4F
// )
// ( tipi delle primitive
#define TIPO_A			0x42	// activate_p
#define TIPO_T			0x43	// terminate_p
#define TIPO_SI			0x44	// sem_ini
#define TIPO_W			0x45	// sem_wait
#define TIPO_S			0x46	// sem_signal
#define TIPO_D			0x49	// delay
#define TIPO_RE			0x4b	// resident
#define TIPO_EP			0x4c	// end_program
#define TIPO_APE		0x52	// activate_pe
#define TIPO_WFI		0x53	// wfi
#define TIPO_FG			0x54	// *fill_gate
#define TIPO_P			0x55	// *panic
#define TIPO_AB			0x56	// *abort_p
#define TIPO_L			0x57	// *log
#define TIPO_TRA		0x58	// trasforma

#define IO_TIPO_HDR		0x62	// readhd_n
#define IO_TIPO_HDW		0x63	// writehd_n
#define IO_TIPO_DMAHDR		0x64	// dmareadhd_n
#define IO_TIPO_DMAHDW		0x65	// dmawritehd_n
#define IO_TIPO_RSEN		0x72	// readse_n
#define IO_TIPO_RSELN		0x73	// readse_ln
#define IO_TIPO_WSEN		0x74	// writese_n
#define IO_TIPO_WSE0		0x75	// writese_0
#define IO_TIPO_RCON		0x76	// readconsole
#define IO_TIPO_WCON		0x77	// writeconsole
#define IO_TIPO_INIC		0x78	// iniconsole

//////////////////////////////////////////////////////////////////////
//                 SCHEDA AUDIO                 					//
//////////////////////////////////////////////////////////////////////
#define IO_TIPO_SET_CS		0x79	// 
#define IO_TIPO_COMMAND_CS		0x80	// 
// * in piu' rispetto al libro
// )

// ( suddivisione della memoria virtuale
//   N    = Numero di entrate in tab4
//   I	  = Indice della prima entrata in tab4
//   SIS  = SIStema
//   MIO  = Modulo IO
//   UTN  = modulo UTeNte
//   C    = Condiviso
//   P    = Privato
#define I_SIS_C		0
#define I_SIS_P		1
#define I_MIO_C		2
#define I_UTN_C       256
#define I_UTN_P	      384

#define N_SIS_C		1
#define N_SIS_P		1
#define N_MIO_C		1
#define N_UTN_C	      128
#define N_UTN_P	      127

// livello fino al quale le tabelle o le pagine delle sezioni condivise devono
// essere precaricate all'avvio. Con 0, che e' il livello delle pagine, le
// sezioni condivise vengono precaricate per intero.  Aumentando il livello
// (fino a 3) si puo' velocizzare l'avvio, al prezzo di causare page fault
// nelle sezioni condivise durante l'esecuzione.
#define PRELOAD_LEVEL	0
// )


////////////////////////////////////////////////////////////////////////////////
//                 SCHEDA AUDIO                 //
////////////////////////////////////////////////////////////////////////////////

//numero registri
#define CS_DIRECT_REG 4
#define CS_REG 32

//costanti per utente
#define STEREO 1
#define MONO 0

#define PLAYBACK 0
#define CAPTURE 1

#define U8 0
#define L16 2

#define START 0
#define PAUSE 1
#define PLAY 2
#define STOP 3

////////////////////////////////////
// bit significativi per maschere//
///////////////////////////////////

//R0 index address register
#define CS_INIT (1<<7) 		// ronly ==1 se si sta inizializzando
#define CS_MCE (1<<6)		// mode change enable
#define CS_TRD (1<<5) 		// ==1 transfer request disable 
#define CS_INDX_MD1 0xf0 	// index for R1
#define CS_INDX_MD2 0xe0	// index for R1

//R2 status ronly
#define CS_INT (1<<0) 		// interrupt status
#define CS_PRDY (1<<1)		// R2 ready for more data
#define CS_PLR (1<<2) 		// data needed for left or right channel
#define CS_PUL (1<<3) 		// data needed for upper or lower byte of the channel
#define CS_SER (1<<4) 		// sample error (I24 indicate the exact sourceof the err)
#define CS_CRDY (1<<5) 		// capture data ready in R3
#define CS_CLR (1<<6) 		// capture data needed for left or right channel
#define CS_CUL (1<<3) 		// capture data needed for upper or lower byte of the channel

//Indici
#define CS_I0 0x00
#define CS_I1 0x01
#define CS_I2 0x02
#define CS_I3 0x03
#define CS_I4 0x04
#define CS_I5 0x05
#define CS_I6 0x06
#define CS_I7 0x07
#define CS_I8 0x08
#define CS_I9 0x09
#define CS_I10 0x0a
#define CS_I11 0x0b
#define CS_I12 0x0c
#define CS_I13 0x0d
#define CS_I14 0x0e
#define CS_I15 0x0f
#define CS_I16 0x10
#define CS_I17 0x11
#define CS_I18 0x12
#define CS_I19 0x13
#define CS_I20 0x14
#define CS_I21 0x15
#define CS_I22 0x16
#define CS_I23 0x17
#define CS_I24 0x18
#define CS_I25 0x19
#define CS_I26 0x1a
#define CS_I27 0x1b
#define CS_I28 0x1c
#define CS_I29 0x1d
#define CS_I30 0x1e
#define CS_I31 0x1f
#define CS_ADDRESS(a,b) ( (a & CS_INDX_MD2) | b )

//Left_ADC_Input_Control I0
#define L_ADC_G 0xf0 		// left adc gain
#define L_MIC_G (1<<5) 		// left mic gain enable
#define L_ADC_S 0x3f 		// left adc input source 

//Right_ADC_Input_Control I1
#define R_ADC_G 0xf0 		// right adc gain
#define R_MIC_G (1<<5) 		// right mic gain enable
#define R_ADC_S 0x3f		// right adc input source 

//Left_AUX1_Input_Control I2
//Right_AUX1_Input_Control I3
//Left_AUX2_Input_Control I4
//Right_AUX2_Input_Control I5

//Left_DAC_Output_Control I6
#define L_DAC_A 0x80		// left dac attenuator
#define CS_LDM (1<<7) 		// left dac mute

//Right_DAC_Output_Control I7
#define R_DAC_A 0x80 		// right dac attenuator
#define CS_RDM (1<<7) 		// right dac mute

 //FS_And_Playback_Data_Format I8
#define CS_C2SL (1<<0) 		// clock 2 source select
#define CS_CFS 0xf1 		// clock frequency divide select
#define CS_SM (1<<4) 		// stereo/mono mode
#define CS_CL 0x1f 			// audio data format

//Interface_Configuration I9
#define CS_PEN (1<<0) 		// playback enable
#define CS_CEN (1<<1) 		// capture enable
#define CS_SDC (1<<2) 		// single dma channel
#define CS_CAL 0xc7 		// calibration
#define CS_PPIO (1<<6) 		// playback data transferred via dma/pio
#define CS_CPIO (1<<7) 		// caputure data transferred via dma/pio

//Pin_Control I10
#define CS_IEN (1<<1) 		// interrupt enable
#define CS_DEN (1<<3) 		// dither enable
#define CS_XCTL 0x3f 		// xctl control

//Error_Status_And_Initialization I11
#define CS_ORL 0x03 		// left overrange detection
#define CS_ORR 0x0c 		// right overrange detection
#define CS_DRS (1<<4) 		// drq status
#define	CS_ACI (1<<5) 		// calibration in progress
#define	CS_PUR (1<<6) 		// playback underrun, data not arrived in time
#define	CS_COR (1<<7) 		// capture overrun	

//MODE_And_ID I12
#define CS_CODEC_ID 0x0a
#define CS_MODE2 (1<<6) 	// mode2 enable

//Loopback_Control I13
#define CS_LBE (1<<0) 		// loopback enable
#define CS_LBA 0x01 		// loopback attenuation

//Playback_Upper_Base_Count I14
//Playback_Lower_Base_Count I15

//Alternate_Feature_Enable_I I16
#define CS_DACZ (1<<0) 		// force the output to AC zero when an underrun error occurs
#define CS_SPE (1<<1) 		// serial port enable
#define CS_SEF 0xf3 		// serial format
#define CS_PMCE (1<<4) 		// playback mode change enable
#define CS_CMCE (1<<5) 		// capture mode change enable
#define CS_TE (1<<6) 		// timer enable
#define CS_OLB (1<<7) 		// output level

//Alternate_Feature_Enable_II I17
#define CS_HPF (1<<0) 		// high pass filter enable
#define CS_XTALE (1<<1) 	// crystal enable
#define CS_APAR (1<<2) 		// playback accumulator reset

//Left_Line_Input_Control i18
#define CS_LEFTLINE 0x80 	// left line mix gain
#define CS_LLM (1<<7) 		// left line mute

//Right_Line_Input_Control I19
#define CS_RIGHTLINE 0x80 	// right line mix gain
#define CS_RLM (1<<7) 		// right line mute

//Timer_Low_Base I20
//Timer_High_Base I21
//RESERVED I22

//Alternate_Feature_Enable_III I23
#define CS_ACF (1<<0) 		// used ehwn pausing a capture stream

//Alternate_Feature_Status I24
#define CS_PU (1<<0) 		// playback underrun, DAC run out of data
#define CS_PO (1<<1) 		// playback overrun
#define CS_CO (1<<2) 		// caputure overrun
#define CS_CU (1<<3) 		// capture underrun
#define CS_PI (1<<4) 		// playback interrupt
#define CS_CI (1<<5) 		// capture interrupt
#define CS_TI (1<<6) 		// timer interrupt

//Version_Chip_ID I25
#define CS_VERSION_NUMBER 0xe0
#define CS_CHIP_ID 0x7

//Mono_Input_And_Output_Control I26
#define CS_MIA 0xe0 		// mono input attenuation
#define CS_MBY (1<<5) 		// mono bypass
#define CS_MOM (1<<6) 		// mono output mute
#define CS_MIM (1<<7) 		// mono input mute

//RESERVED_2 I27

//Capture_Data_Format I28
#define CS_SM (1<<4) 		// stereo mono select
#define CS_CDF 0xe0 		// capture data format

//RESERVED_3 I29
//Capture_Upper_Base_Count I30
//Capture_Lower_Base_Count I31
