// io.cpp
//
#include "costanti.h"
#include "libce.h"

//#define BOCHS
////////////////////////////////////////////////////////////////////////////////
//    COSTANTI                                                                //
////////////////////////////////////////////////////////////////////////////////


const natl PRIO = 1000;
const natl LIV = LIV_SISTEMA;


////////////////////////////////////////////////////////////////////////////////
//                        CHIAMATE DI SISTEMA USATE                           //
////////////////////////////////////////////////////////////////////////////////

extern "C" natl activate_pe(void f(int), int a, natl prio, natl liv, natb type);
extern "C" void terminate_p();
extern "C" void sem_wait(natl sem);
extern "C" void sem_signal(natl sem);
extern "C" natl sem_ini(int val);
extern "C" void wfi();	//
extern "C" void abort_p();
extern "C" addr trasforma(addr ff);
extern "C" void panic(const char *msg);

void *mem_alloc(natl dim);
void mem_free(void *p);

////////////////////////////////////////////////////////////////////////////////
//                    GESTIONE DELLE INTERFACCE SERIALI                 //
////////////////////////////////////////////////////////////////////////////////

enum funz { input_n, input_ln, output_n, output_0 };  //

struct interfse_reg {	//
	ioaddr iRBR, iTHR, iLSR, iIER, iIIR;
};

struct des_se {		//
	interfse_reg indreg;
	natl mutex;
	natl sincr;
	natl cont;
	addr punt;
	funz funzione;
	natb stato;
};

const natl S = 2;
des_se com[S] = {
	{	// com[0]
		{	// indreg
			0x03f8,	// iRBR
			0x03f8,	// iTHR
			0x03fd,	// iLSR
			0x03f9,	// iIER
			0x03fa,	// iIIR
		},
		0,	// mutex
		0,	// sincr
		0,	// cont
		0,	// punt
		(funz)0,// funzione
		0,	// stato
	},
	{	// com[0]
		{	// indreg
			0x02f8,	// iRBR
			0x02f8,	// iTHR
			0x03fd,	// iLSR
			0x02f9,	// iIER
			0x02fa,	// iIIR
		},
		0,	// mutex
		0,	// sincr
		0,	// cont
		0,	// punt
		(funz)0,// funzione
		0,	// stato
	}
};

void input_com(des_se* p_des);	//
void output_com(des_se* p_des);	//
void estern_com(int i) //
{
	natb r;
	des_se *p_des;
	p_des = &com[i];
	for(;;) {
		inputb(p_des->indreg.iIIR, r);
		if ((r&0x06) == 0x04)
			input_com(p_des);
		else if ((r&0x06) == 0x02)
			output_com(p_des);
		wfi();
	}
}

void startse_in(des_se *p_des, natb vetti[], natl quanti, funz op); // [9.2.1]
extern "C" void c_readse_n(natl serial, natb vetti[], natl quanti, natb& errore) // [9.2.1]
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "readse_n con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_in(p_des, vetti, quanti, input_n);
	sem_wait(p_des->sincr);
	errore = p_des->stato;
	sem_signal(p_des->mutex);
}

extern "C" void c_readse_ln(natl serial, natb vetti[], int& quanti, natb& errore)
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "readse_ln con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_in(p_des, vetti, 80, input_ln);
	sem_wait(p_des->sincr);
	quanti = p_des->cont;
	errore = p_des->stato;
	sem_signal(p_des->mutex);
}

extern "C" void go_inputse(ioaddr i_ctr);
void startse_in(des_se *p_des, natb vetti[], natl quanti, funz op) // [9.2.1]
{
	p_des->cont = quanti;
	p_des->punt = vetti;
	p_des->funzione = op;
	go_inputse(p_des->indreg.iIER);
}

extern "C" void halt_inputse(ioaddr i_ctr);
void input_com(des_se *p_des) // [9.2.1]
{
	natb c; bool fine;
	fine = false;

	halt_inputse(p_des->indreg.iIER);

	inputb(p_des->indreg.iLSR, c);

	p_des->stato = c & 0x1e;
	if (p_des->stato != 0)
		fine = true;
	else {
		inputb(p_des->indreg.iRBR, c);
		if (p_des->funzione == input_n) {
			*static_cast<natb*>(p_des->punt) = c; // memorizzazione
			p_des->punt = static_cast<natb*>(p_des->punt) + 1;
			p_des->cont--;
			if(p_des->cont == 0)
				fine = true;
		} else {
			if ( (p_des->funzione == input_ln) ) {
				if(c == '\r' || c == '\n') {
					fine = true;
					p_des->cont = 80 - p_des->cont;
				} else {
					*static_cast<natb*>(p_des->punt) = c; // memorizzazione
					p_des->punt = static_cast<natb*>(p_des->punt) + 1;
					p_des->cont--;
					if (p_des->cont == 0) {
						fine = true;
						p_des->cont = 80;
					}
				}
			}
		}
	}

	if(fine == true) {
		*static_cast<natb*>(p_des->punt) = 0;	// carattere nullo
		sem_signal(p_des->sincr);
	} else
		go_inputse(p_des->indreg.iIER);
}

void startse_out(des_se *p_des, natb vetto[], natl quanti, funz op);
extern "C" void c_writese_n(natl serial, natb vetto[], natl quanti)	// [9.2.2]
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "writese_n con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_out(p_des, vetto, quanti, output_n);
	sem_wait(p_des->sincr);
	sem_signal(p_des->mutex);
}

extern "C" void c_writese_0(natl serial, natb vetto[], natl &quanti)
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "writese_0 con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_out(p_des, vetto, 0, output_0);
	sem_wait(p_des->sincr);
	quanti = p_des->cont;
	sem_signal(p_des->mutex);
}

extern "C" void go_outputse(ioaddr i_ctr);
void startse_out(des_se *p_des, natb vetto[], natl quanti, funz op) // [9.2.2]
{
	p_des->cont = quanti;
	p_des->punt = vetto;
	p_des->funzione = op;
	go_outputse(p_des->indreg.iIER);
	output_com(p_des);
}

extern "C" void halt_outputse(ioaddr i_ctr);
void output_com(des_se *p_des)	// [9.2.2]
{
	natb c; bool fine;
        fine = false;

	if (p_des->funzione == output_n) {
		p_des->cont--;
		if(p_des->cont == 0) {
			fine = true;
			halt_outputse(p_des->indreg.iIER);
		}
		c = *static_cast<natb*>(p_des->punt); //prelievo
		outputb(c, p_des->indreg.iTHR);
		p_des->punt = static_cast<natb*>(p_des->punt) + 1;
	} else if (p_des->funzione == output_0) {
		c = *static_cast<natb*>(p_des->punt); //prelievo
		if (c == 0) {
			fine = true;
			halt_outputse(p_des->indreg.iIER);
		} else {
			outputb(c, p_des->indreg.iTHR);
			p_des->cont++;
			p_des->punt = static_cast<natb*>(p_des->punt) + 1;
		}
	}

	if (fine == true)
		sem_signal(p_des->sincr);

}

// ( inizializzazione delle interfacce seriali
extern "C" void com_setup(void);	// vedi "io.S"
// interruzioni hardware delle interfacce seriali
int com_irq[S] = { 4, 3 };

bool com_init()
{
	des_se *p_des;
	natl id;
	natl i, com_base_prio = PRIO;

	com_setup();

	for(i = 0; i < S; ++i) {
		p_des = &com[i];

		if ( (p_des->mutex = sem_ini(1)) == 0xFFFFFFFF) {
			flog(LOG_ERR, "com: impossibile creare mutex");
			return false;
		}
		if ( (p_des->sincr = sem_ini(0)) == 0xFFFFFFFF) {
			flog(LOG_ERR, "com: impossibile creare sincr");
			return false;
		}

		id = activate_pe(estern_com, i, com_base_prio - i, LIV, com_irq[i]);
		if (id == 0xFFFFFFFF) {
			flog(LOG_ERR, "com: impossibile creare proc. esterno");
			return false;
		}

	}
	flog(LOG_INFO, "com: inizializzate %d seriali", S);
	return true;
}
// )

////////////////////////////////////////////////////////////////////////////////
//                         GESTIONE DELLA CONSOLE                       //
////////////////////////////////////////////////////////////////////////////////

const natl COLS = 80; 	//
const natl ROWS = 25;	//
const natl VIDEO_SIZE = COLS * ROWS;	//

struct interfvid_reg {	//
	ioaddr iIND, iDAT;
};

struct des_vid {	//
	interfvid_reg indreg;
	natw* video;
	natl x, y;
	natw attr;
};

const natl MAX_CODE = 40; //
struct interfkbd_reg {	//
	ioaddr iRBR, iTBR, iCMR, iSTR;
};

struct des_kbd { //
	interfkbd_reg indreg;
	addr punt;
	natl cont;
	bool shift;
	natb tab[MAX_CODE];
	natb tabmin[MAX_CODE];
	natb tabmai[MAX_CODE];
};

struct des_console { //
	natl mutex;
	natl sincr;
	des_kbd kbd;
	des_vid vid;
};

des_console console = {
	0,	// mutex (da inizializzare)
	0,	// sync  (da inizializzare)
	{	// kbd
		{	// indreg
			0x60,	// iRBR
			0x60,	// iTBR
			0x64,	// iCMR
			0x64,	// iSTR
		},
		0,	// punt
		0,	// cont
		0,	// shift
		{	// tab
			0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b,
			0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19,
			0x1E, 0x1F, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25,
			0x26, 0x2C, 0x2D, 0x2E, 0x2F, 0x30, 0x31, 0x32,
			0x39, 0x1C, 0x0e, 0x01
		},
		{	// tamin
			'1', '2', '3', '4', '5', '6', '7', '8', '9', '0',
			'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p',
			'a', 's', 'd', 'f', 'g', 'h', 'j', 'k',
			'l', 'z', 'x', 'c', 'v', 'b', 'n', 'm',
			' ', '\n', '\b', 0x1B
		},
		{	// tabmai
			'!', '"', '@', '$', '%', '&', '/', '(', ')', '=',
			'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P',
			'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K',
			'L', 'Z', 'X', 'C', 'V', 'B', 'N', 'M',
			' ', '\r', '\b', 0x1B
		}
	},
	{	// vid
		{	// indreg
			0x03d4,	// iIND
			0x03d5,	// iDAT
		},
		(natw*)0xb8000,	// video
		0,		// x
		0,		// y
		0x4b00,		// attr
	}
};

extern "C" void cursore(ioaddr iIND, ioaddr iDAT, int x, int y); //

void scroll(des_vid *p_des)	//
{
	for (natl i = 0; i < VIDEO_SIZE - COLS; i++)
		p_des->video[i] = p_des->video[i + COLS];
	for (natl i = 0; i < COLS; i++)
		p_des->video[VIDEO_SIZE - COLS + i] = p_des->attr | ' ';
	p_des->y--;
}

void writeelem(natb c) {	//
	des_vid* p_des = &console.vid;
	switch (c) {
	case 0:
		break;
	case '\r':
		p_des->x = 0;
		break;
	case '\n':
		p_des->x = 0;
		p_des->y++;
		if (p_des->y >= ROWS)
			scroll(p_des);
		break;
	case '\b':
		if (p_des->x > 0 || p_des->y > 0) {
			if (p_des->x == 0) {
				p_des->x = COLS - 1;
				p_des->y--;
			} else
				p_des->x--;
		}
		break;
	default:
		p_des->video[p_des->y * COLS + p_des->x] = p_des->attr | c;
		p_des->x++;
		if (p_des->x >= COLS) {
			p_des->x = 0;
			p_des->y++;
		}
		if (p_des->y >= ROWS)
			scroll(p_des);
		break;
	}
	cursore(p_des->indreg.iIND, p_des->indreg.iDAT,
		p_des->x, p_des->y);
}

void writeseq(cstr seq)	//
{
	const natb* pn = static_cast<const natb*>(seq);
	while (*pn != 0) {
		writeelem(*pn);
		pn++;
	}
}

extern "C" void c_writeconsole(cstr buff) //
{
	des_console *p_des = &console;
	sem_wait(p_des->mutex);
#ifndef AUTOCORR
	writeseq(buff);
	writeelem('\n');
#else /* AUTOCORR */
	flog(LOG_USR, "%s", buff);
#endif /* AUTOCORR */
	sem_signal(p_des->mutex);
}

extern "C" void go_inputkbd(interfkbd_reg* indreg); //
extern "C" void halt_inputkbd(interfkbd_reg* indreg); //

void startkbd_in(des_kbd* p_des, str buff, natl dim)
{
	p_des->punt = buff;
	p_des->cont = dim;
	go_inputkbd(&p_des->indreg);
}

extern "C" void c_readconsole(str buff, natl& quanti) //
{
	des_console *p_des;

	p_des = &console;
	sem_wait(p_des->mutex);
	startkbd_in(&p_des->kbd, buff, quanti);
	sem_wait(p_des->sincr);
	quanti = p_des->kbd.cont;
	sem_signal(p_des->mutex);
}

natb converti(des_kbd* p_des, natb c) { //
	natb cc;
	natl pos = 0;
	while (pos < MAX_CODE && p_des->tab[pos] != c)
		pos++;
	if (pos == MAX_CODE)
		return 0;
	if (p_des->shift)
		cc = p_des->tabmai[pos];
	else
		cc = p_des->tabmin[pos];
	return cc;
}

void estern_kbd(int h) //
{
	des_console *p_des = &console;
	natb a, c;
	bool fine;

	for(;;) {
		halt_inputkbd(&p_des->kbd.indreg);

		inputb(p_des->kbd.indreg.iRBR, c);

		fine = false;
		switch (c) {
		case 0x2a: // left shift make code
			p_des->kbd.shift = true;
			break;
		case 0xaa: // left shift break code
			p_des->kbd.shift = false;
			break;
		default:
			if (c < 0x80) {
				a = converti(&p_des->kbd, c);
				if (a == 0)
					break;
				if (a == '\b') {
					if (p_des->kbd.cont < 80) {
						p_des->kbd.punt = static_cast<natb*>(p_des->kbd.punt) - 1;
						p_des->kbd.cont++;
						writeseq("\b \b");
					}
				} else if (a == '\r' || a == '\n') {
					fine = true;
					p_des->kbd.cont = 80 - p_des->kbd.cont;
					*static_cast<natb*>(p_des->kbd.punt) = 0;
					writeseq("\r\n");
				} else {
					*static_cast<natb*>(p_des->kbd.punt) = a;
					p_des->kbd.punt = static_cast<natb*>(p_des->kbd.punt) + 1;
					p_des->kbd.cont--;
					writeelem(a);
					if (p_des->kbd.cont == 0) {
						fine = true;
						p_des->kbd.cont = 80;
					}
				}
			}
			break;
		}
		if (fine == true)
			sem_signal(p_des->sincr);
		else
			go_inputkbd(&p_des->kbd.indreg);
		wfi();
	}
}

// (* inizializzazioni
extern "C" void abilita_tastiera(void);
bool vid_init();

extern "C" void c_iniconsole(natb cc)
{
	des_vid *p_des = &console.vid;
	p_des->attr = static_cast<natw>(cc) << 8;
	vid_init();
}

// Interruzione hardware della tastiera
const int KBD_IRQ = 1;

bool kbd_init()
{
	// blocchiamo subito le interruzioni generabili dalla tastiera
	halt_inputkbd(&console.kbd.indreg);

	if (activate_pe(estern_kbd, 0, PRIO, LIV, KBD_IRQ) == 0xFFFFFFFF) {
		flog(LOG_ERR, "kbd: impossibile creare estern_kbd");
		return false;
	}
	return true;
}

extern "C" des_vid vid;

bool vid_init()
{
	des_vid *p_des = &console.vid;
	for (natl i = 0; i < VIDEO_SIZE; i++)
		p_des->video[i] = p_des->attr | ' ';
	cursore(p_des->indreg.iIND, p_des->indreg.iDAT,
		p_des->x, p_des->y);
	flog(LOG_INFO, "vid: video inizializzato");
	return true;
}

bool console_init()
{
	des_console *p_des = &console;

	if ( (p_des->mutex = sem_ini(1)) == 0xFFFFFFFF) {
		flog(LOG_ERR, "kbd: impossibile creare mutex");
		return false;
	}
	if ( (p_des->sincr = sem_ini(0)) == 0xFFFFFFFF) {
		flog(LOG_ERR, "kbd: impossibile creare sincr");
		return false;
	}
	return kbd_init() && vid_init();
}

// *)

// inerfacce ATA

enum hd_cmd { WRITE_SECT = 0x30, READ_SECT = 0x20, WRITE_DMA = 0xCA, READ_DMA = 0xC8 };
struct interfata_reg {
	ioaddr iBR;
	ioaddr iCNL, iCNH, iSNR, iHND, iSCR, iERR,
	       iCMD, iSTS, iDCR, iASR;
};
struct pci_ata
{
	ioaddr iBMCMD, iBMSTR, iBMDTPR;
};
struct des_ata {
	interfata_reg indreg;
	pci_ata bus_master;
	natl prd[2];
	hd_cmd comando;
	natb errore;
	natl mutex;
	natl sincr;
	natb cont;
	addr punt;
};
des_ata hd = {
	{
		0x0170,	// iBR
		0x0174, // iCNL
		0x0175, // iCNH
		0x0173, // iSNR
		0x0176, // iHND
		0x0172, // iSCR
		0x0171, // iERR
		0x0177, // iCMD
		0x0177, // iSTS
		0x0376, // iDCR
		0x0376  // iASR
	}
	// il resto e' inizializzato a zero
};


const natb HD_IRQ = 15;

extern "C" void hd_write_address(des_ata* p_des, natl primo);
// scrive primo nei registri di indirizzo CNL, CNH, SNR e HND
extern "C" void hd_write_command(hd_cmd cmd, ioaddr iCMD);
// scrive cmd nel registro iCMD e aspetta che BSY torni a 0

extern "C" bool hd_wait_data(ioaddr iSTS) // [9.3.1]
{
	natb stato;

	do {
		inputb(iSTS, stato);
	} while (stato & 0x80);

	return ( (stato & 0x08) && !(stato & 0x01) );
}
// attende che BSY passi a zero e DRQ a 1. Restituisce false se ERR vale 1
extern "C" void hd_go_inout(ioaddr iSTS);
// abilita l'interfaccia a generare interruzioni
extern "C" void hd_halt_inout(ioaddr iSTS);
// disabilita l'interfaccia a generare interruzioni

void hd_componi_prd(des_ata* p_dmades, addr iff, natw quanti)
{
	p_dmades->prd[0] = reinterpret_cast<natq>(iff);
	p_dmades->prd[1] = 0x80000000 | quanti;					// EOT posto a 1
}
extern "C" void hd_select_device(short ms, ioaddr iHND);
void hd_sel_drv(des_ata* p_des) //
{
	natb stato;

	hd_select_device(0, p_des->indreg.iHND);

	do {
		inputb(p_des->indreg.iSTS, stato);
	} while ( (stato & 0x80) || (stato & 0x08) );
}
void starthd_in(des_ata *p_des, natw vetti[], natl primo, natb quanti);
extern "C" void c_readhd_n(natw vetti[], natl primo,
		natb quanti, natb &errore)
{
	des_ata *p_des;
	p_des = &hd;
	sem_wait(p_des->mutex);
	starthd_in(p_des, vetti, primo, quanti);
	sem_wait(p_des->sincr);
	errore = p_des->errore;
	sem_signal(p_des->mutex);
}
void starthd_out(des_ata *p_des, natw vetto[], natl primo, natb quanti);
extern "C" void c_writehd_n(natw vetto[], natl primo,
		natb quanti, natb &errore)
{
	des_ata *p_des;
	p_des = &hd;
	sem_wait(p_des->mutex);
	starthd_out(p_des, vetto, primo, quanti);
	sem_wait(p_des->sincr);
	errore = p_des->errore;
	sem_signal(p_des->mutex);
}
void starthd_in(des_ata *p_des, natw vetti[], natl primo, natb quanti)
{
	p_des->cont = quanti;
	p_des->punt = vetti;
	p_des->comando = READ_SECT;
	hd_sel_drv(p_des);
	hd_write_address(p_des, primo);
	outputb(quanti, p_des->indreg.iSCR);
	hd_go_inout(p_des->indreg.iDCR);
	hd_write_command(READ_SECT, p_des->indreg.iCMD);
}
void starthd_out(des_ata *p_des, natw vetto[], natl primo, natb quanti)
{
	p_des->cont = quanti;
	p_des->punt = vetto + DIM_BLOCK / 2;
	p_des->comando = WRITE_SECT;
	hd_sel_drv(p_des);
	hd_write_address(p_des, primo);
	outputb(quanti,p_des->indreg.iSCR);
	hd_go_inout(p_des->indreg.iDCR);
	hd_write_command(WRITE_SECT, p_des->indreg.iCMD);
	hd_wait_data(p_des->indreg.iSTS);
	outputbw(vetto, DIM_BLOCK/2, p_des->indreg.iBR);
}
void dmastarthd_in(des_ata *p_des, natw vetti[], natl primo, natb quanti);
extern "C" void c_dmareadhd_n(natw vetti[], natl primo, natb quanti,
		natb &errore)
{
	des_ata *p_des;
	p_des = &hd;
	sem_wait(p_des->mutex);
	dmastarthd_in(p_des, vetti, primo, quanti);
	sem_wait(p_des->sincr);
	errore = p_des->errore;
	sem_signal(p_des->mutex);
}
void dmastarthd_out(des_ata *p_des, natw vetto[], natl primo, natb quanti);
extern "C" void c_dmawritehd_n(natw vetto[], natl primo, natb quanti,
		natb& errore)
{
	des_ata *p_des;
	p_des = &hd;
	sem_wait(p_des->mutex);
	dmastarthd_out(p_des, vetto, primo, quanti);
	sem_wait(p_des->sincr);
	errore = p_des->errore;
	sem_signal(p_des->mutex);
}
void dmastarthd_in(des_ata *p_des, natw vetti[], natl primo, natb quanti)
{
	// la scrittura ini iBMDTPR di &prd[0] avviene in fase di inizializzazione
	natb work; addr iff;
	p_des->comando = READ_DMA;
	p_des->cont = 1;					// informazione per il driver
	iff = trasforma(vetti);
	hd_componi_prd(p_des, iff, quanti * DIM_BLOCK);
	inputb(p_des->bus_master.iBMCMD, work);
	work |= 0x08;							// lettura da PCI
	outputb(work, p_des->bus_master.iBMCMD);
	inputb(p_des->bus_master.iBMSTR, work);
	work &= 0xF9;						// bit interruzione ed errore a 0
	outputb(work, p_des->bus_master.iBMSTR);
	hd_sel_drv(p_des);
	hd_write_address(p_des, primo);
	outputb(quanti, p_des->indreg.iSCR);
	hd_go_inout(p_des->indreg.iDCR);
	hd_write_command(READ_DMA, p_des->indreg.iCMD);
	inputb(p_des->bus_master.iBMCMD, work);
	work |= 0x01; 							// avvio dell’operazione
	outputb(work, p_des->bus_master.iBMCMD);
}
void dmastarthd_out(des_ata *p_des, natw vetto[], natl primo, natb quanti)
{
	// la scrittura in iBMDTPR di &prd[0] avviene in fase di inizializzazione
	natb work; addr iff;
	p_des->comando = WRITE_DMA;
	p_des->cont = 1; 				// informazione per il driver
	iff = trasforma(vetto);
	hd_componi_prd(p_des, iff, quanti * DIM_BLOCK);
	inputb(p_des->bus_master.iBMCMD, work);
	work &= 0xF7; 					// scrittura verso PCI
	outputb(work, p_des->bus_master.iBMCMD);
	inputb(p_des->bus_master.iBMSTR, work);
	work &= 0xF9;					// bit interruzione ed errore a 0
	outputb(work, p_des->bus_master.iBMSTR);
	hd_sel_drv(p_des);
	hd_write_address(p_des, primo);
	outputb(quanti, p_des->indreg.iSCR);
	hd_go_inout(p_des->indreg.iDCR);
	hd_write_command(WRITE_DMA, p_des->indreg.iCMD);
	inputb(p_des->bus_master.iBMCMD, work);
	work |= 1;								// avvio dell’operazione
	outputb(work, p_des->bus_master.iBMCMD);
}

void esternAta(int h)			// codice commune ai 2 processi esterni ATA
{
	des_ata* p_des = &hd;
	natb stato, work;
	for(;;) {
		p_des->cont--;
		if (p_des->cont == 0)
			hd_halt_inout(p_des->indreg.iDCR);
		p_des->errore = 0;
		inputb(p_des->indreg.iSTS, stato); 				// ack dell'interrupt
		switch (p_des->comando) {
		case READ_SECT:
			if (!hd_wait_data(p_des->indreg.iSTS))
				inputb(p_des->indreg.iERR, p_des->errore);
			else
				inputbw(p_des->indreg.iBR, static_cast<natw*>(p_des->punt),
						DIM_BLOCK / 2);
			p_des->punt = static_cast<natw*>(p_des->punt) + DIM_BLOCK / 2;
			break;
		case WRITE_SECT:
			if (p_des->cont != 0) {
				if (!hd_wait_data(p_des->indreg.iSTS))
					inputb(p_des->indreg.iERR, p_des->errore);
				else
					outputbw(static_cast<natw*>(p_des->punt),
							DIM_BLOCK / 2, p_des->indreg.iBR);
				p_des->punt = static_cast<natw*>(p_des->punt) +
					DIM_BLOCK / 2;
			}
			break;
		case READ_DMA:
		case WRITE_DMA:
			inputb(p_des->bus_master.iBMCMD, work);
			work &= 0xFE;				// azzeramento del bit n. 0 (start/stop)
			outputb(work, p_des->bus_master.iBMCMD);
			inputb(p_des->bus_master.iBMSTR, stato);	// ack  interrupt in DMA
			if ((stato & 0x05) == 0)
				inputb(p_des->indreg.iERR, p_des->errore);
			break;
		}
		if (p_des->cont == 0)
			sem_signal(p_des->sincr);
		wfi();
	}
}

bool hd_init()
{
	natl id;
	natb bus = 0, dev = 0, fun = 0;
	natb code[] = { 0xff, 0x01, 0x01 };
	des_ata* p_des;

	p_des = &hd;

	if ( (p_des->mutex = sem_ini(1)) == 0xFFFFFFFF) {
		flog(LOG_ERR, "hd: impossibile creare mutex");
		return false;
	}
	if ( (p_des->sincr = sem_ini(0)) == 0xFFFFFFFF) {
		flog(LOG_ERR, "hd: impossibile creare sincr");
		return false;
	}

	if (!pci_find_class(bus, dev, fun, code)) {
		flog(LOG_WARN, "hd: bus master non trovato");
	} else {
		natb prog_if = pci_read_confb(bus, dev, fun, 0x9);
		if (prog_if & 0x80) {
			natl base = pci_read_confl(bus, dev, fun, 0x20);
			base &= ~0x1;
			hd.bus_master.iBMCMD  = (ioaddr)(base + 0x08);
			hd.bus_master.iBMSTR  = (ioaddr)(base + 0x0a);
			hd.bus_master.iBMDTPR = (ioaddr)(base + 0x0c);
			addr iff = trasforma(&hd.prd[0]);
			outputl(reinterpret_cast<natq>(iff),
					hd.bus_master.iBMDTPR);
			natw cmd = pci_read_confw(bus, dev, fun, 4);
			pci_write_confw(bus, dev, fun, 4, cmd | 0x5);
		}
	}

	id = activate_pe(esternAta, 0, PRIO, LIV, HD_IRQ);
	if (id == 0xFFFFFFFF) {
		flog(LOG_ERR, "com: impossibile creare proc. esterno");
		return false;
	}
	return true;
}


natl mem_mutex;

void* mem_alloc(natl dim)
{
	void *p;

	sem_wait(mem_mutex);
	p = alloca(dim);
	sem_signal(mem_mutex);

	return p;
}


void mem_free(void* p)
{
	sem_wait(mem_mutex);
	dealloca(p);
	sem_signal(mem_mutex);
}
//////////////////////////////////////////////////////////////////////
//                 SCHEDA AUDIO                 					//
//////////////////////////////////////////////////////////////////////

///////////////////////	STRUTTURE DATI ///////////////////////////////

//indici registri diretti 
enum {
    Index_Address,
    Index_Data,
    Status,
    PIO_Data
};
//indici registri indiretti
enum {
    Left_ADC_Input_Control,
    Right_ADC_Input_Control,
    Left_AUX1_Input_Control,
    Right_AUX1_Input_Control,
    Left_AUX2_Input_Control,
    Right_AUX2_Input_Control,
    Left_DAC_Output_Control,
    Right_DAC_Output_Control,
    FS_And_Playback_Data_Format,
    Interface_Configuration,
    Pin_Control,
    Error_Status_And_Initialization,
    MODE_And_ID,
    Loopback_Control,
    Playback_Upper_Base_Count,
    Playback_Lower_Base_Count,
    Alternate_Feature_Enable_I,
    Alternate_Feature_Enable_II,
    Left_Line_Input_Control,
    Right_Line_Input_Control,
    Timer_Low_Base,
    Timer_High_Base,
    RESERVED,
    Alternate_Feature_Enable_III,
    Alternate_Feature_Status,
    Version_Chip_ID,
    Mono_Input_And_Output_Control,
    RESERVED_2,
    Capture_Data_Format,
    RESERVED_3,
    Capture_Upper_Base_Count,
    Capture_Lower_Base_Count
};
//reg dma
enum 
{
	Start_Address_Register,
	Count_Register,
	Single_Channel_Mask_Register,
	Mode_Register,
	FlipFlop_Reset_Register,
	Page_Register
};

//struttura usata per i registri diretti 
struct registro
{
	ioaddr indreg;
	natb valore;
};

struct CScard
{
	registro direct_reg[CS_DIRECT_REG]; //vettore per i registri diretti (add e val)
	natb reg[CS_REG]; //indiretti salvati in memoria
	
	natb irq; // numero piedino di interruzione

	natl CS_mutex; //mutex per dispositivo

	bool presente;
	bool set;
};

CScard cs_card;

struct suono 
{
	char* buffer;
	int quanti;
	natl mutex;
	natl sync;
};

suono cs_suono;
char* sysBuff;


struct dmacCH3
{
	ioaddr dmac_reg[6];
	natl mutex;
};

dmacCH3 dmac;

///////////////////////	FUNZIONI UTILITA ///////////////////////////////

void set_Index_Address(natb index)
{
	//non serve mutex poiche attivato dal chiamante
	CScard* p_des;
	p_des = &cs_card;
	p_des->direct_reg[Index_Address].valore = CS_ADDRESS(p_des->direct_reg[Index_Address].valore, index);
	outputb(p_des->direct_reg[Index_Address].valore,p_des->direct_reg[Index_Address].indreg);
}

void stampaReg()
{
	CScard* p_des;
	p_des = &cs_card;

	natb ctx;

	for (natb i = 0; i < 4; ++i)
	{
		inputb(p_des->direct_reg[i].indreg,ctx);
		flog(LOG_INFO," Registro diretto %d : %x",i,(ctx & 0xff));
	}

	for (natb i = 0; i < CS_REG; ++i)
	{
		set_Index_Address(i);
		inputb(p_des->direct_reg[Index_Data].indreg,ctx);
		flog(LOG_INFO," Registro %d : %x",i,(ctx & 0xff));
	}
}

void set_dmac(addr add,natb low_count,natb high_count)
{
	//non serve mutex poiche attivato dal chiamante
	dmacCH3* p_dmac;
	p_dmac = &dmac;
	natb* address = (natb*)&add;
	natb low_add = address[0];
	natb high_add = address[1];
	natb page = address[2];
	//flog(LOG_INFO,"indirizzi tripli : indirizzo fisico:%x ,page %x , high %x , low%x",add,address[2],address[1],address[0]);
	//flog(LOG_INFO,"quanti  : upper %x , lower %x",high_count,low_count);
	outputb(0x07,p_dmac->dmac_reg[Single_Channel_Mask_Register]); //maschero drq su ch3
	outputb(0xFF,p_dmac->dmac_reg[FlipFlop_Reset_Register]); //resetto master flip flop per doppia scrittura
	outputb(low_add,p_dmac->dmac_reg[Start_Address_Register]); //low byte
	outputb(high_add,p_dmac->dmac_reg[Start_Address_Register]); //high byte
	outputb(0xFF,p_dmac->dmac_reg[FlipFlop_Reset_Register]); //resetto master flipflop
	outputb(low_count,p_dmac->dmac_reg[Count_Register]); //low
	outputb(high_count,p_dmac->dmac_reg[Count_Register]); //high
	outputb(page,p_dmac->dmac_reg[Page_Register]); //total address
	outputb(0x4b,p_dmac->dmac_reg[Mode_Register]); //single trnasfer,addr inc,no autoinit(not supp),read,ch3 = 4b
	outputb(0x03,p_dmac->dmac_reg[Single_Channel_Mask_Register]); //tolgo maschera a drq	
}

void set_pen(bool a)
{	// attivo pen 
	CScard* p_des;
	p_des = &cs_card;
	sem_wait(p_des->CS_mutex);
	set_Index_Address(CS_I9);
	p_des->reg[Interface_Configuration] = (a == 1)? (p_des->reg[Interface_Configuration] |CS_PEN) : (p_des->reg[Interface_Configuration] & (~CS_PEN));
	outputb(p_des->reg[Interface_Configuration],p_des->direct_reg[Index_Data].indreg);
	sem_signal(p_des->CS_mutex);
}

void set_lowerbase(natb val)
{	//scrivo lower base per dma
	CScard* p_des;
	p_des = &cs_card;
	sem_wait(p_des->CS_mutex);	
	set_Index_Address(CS_I15);
	p_des->direct_reg[Index_Data].valore = val;
	outputb(p_des->direct_reg[Index_Data].valore,p_des->direct_reg[Index_Data].indreg);
	sem_signal(p_des->CS_mutex);
}	

void set_upperbase(natb val)
{	//scrivo in upper base per dma
	CScard* p_des;
	p_des = &cs_card;
	sem_wait(p_des->CS_mutex);	
	set_Index_Address(CS_I14);
	p_des->direct_reg[Index_Data].valore = val;
	outputb(p_des->direct_reg[Index_Data].valore,p_des->direct_reg[Index_Data].indreg);
	sem_signal(p_des->CS_mutex);
}

//settaggio bit per cambiare configurazione scheda
void set_mce()
{	
	CScard* p_des;
	p_des = &cs_card;	
	p_des->direct_reg[Index_Address].valore |= CS_MCE;
	outputb(p_des->direct_reg[Index_Address].valore,p_des->direct_reg[Index_Address].indreg);
}

void set_pmce()
{
	CScard* p_des;
	p_des = &cs_card;
	set_Index_Address(CS_I16);
	outputb(0x10,p_des->direct_reg[Index_Data].indreg);
}

void reset_mce()
{
	CScard* p_des;
	p_des = &cs_card;	
	p_des->direct_reg[Index_Address].valore &= (~CS_MCE);
	outputb(p_des->direct_reg[Index_Address].valore,p_des->direct_reg[Index_Address].indreg);
}

void reset_pmce()
{
	CScard* p_des;
	p_des = &cs_card;
	set_Index_Address(CS_I16);
	outputb(0x00,p_des->direct_reg[Index_Data].indreg);
}
//////////////////////////	CORPO ///////////////////////////////

bool CS_found(natw base)
{
	bool found = true;
	natb r0,r1;
	ioaddr indR0,indR1;
	indR0 = (ioaddr)(base+0);
	indR1 = (ioaddr)(base+1);
	inputb(indR0,r0);
	if ( ( (r0 & CS_INIT) == CS_INIT ) ) //controllo se la scheda si sta inizializzando e non puo essere letta
	{	
		flog(LOG_ERR,"cs4231a non pronta");
		return !found;
	}
	r0 = ( (r0 & CS_INDX_MD1) | CS_I12 ); 
	outputb(r0,indR0);	//indirizzo il registro I12
	inputb(indR1,r1);
	if ( (r1 & 0x0f) != CS_CODEC_ID) //controllo se l'id combacia
		return !found;
	else
		return found;

}

void suono_init(char* buff,int quanti)
{
	suono* p_suono;
	p_suono = &cs_suono;	
	sem_wait(p_suono->mutex);
	p_suono->buffer = buff; //virtuale di sys
	p_suono->quanti = quanti; //byte
	sem_signal(p_suono->mutex);
}

extern "C" bool c_set_cs(char* buf,int quanti,bool mod,bool p_c,natb format,int khz )
{
	CScard* p_des;
	p_des = &cs_card;

	if (format != 0 && format != 2 )
	{
		c_writeconsole("Selezionare un formato valido fra U8, L16");
		p_des->set = false;
		return false;				
	}
	
	if (quanti<=0)
	{
		c_writeconsole("Il numero di byte deve essere maggiore di zero");
		p_des->set = false;
		return false;
	}
	sysBuff = (char*)mem_alloc(quanti*sizeof(char)); 
	if(!sysBuff)
	{
		flog(LOG_ERR,"Impossibile allocare buffer");
		p_des->set = false;
		return false;	
	}
	flog(LOG_INFO," Allocato buffer in i/o condiviso. Indirizzo virtuale %lx dim %d",sysBuff,quanti);
	memcpy(sysBuff,buf,quanti);

	suono_init(sysBuff,quanti);

	sem_wait(p_des->CS_mutex);	

	//abilito int
	p_des->reg[Pin_Control] |= CS_IEN;
	set_Index_Address(CS_I10);
	outputb(p_des->reg[Pin_Control],p_des->direct_reg[Index_Data].indreg);

	set_mce();
	set_pmce();
	//setto modalita 
	if (mod == STEREO)
		p_des->reg[FS_And_Playback_Data_Format] |= CS_SM;
	else
		p_des->reg[FS_And_Playback_Data_Format] &= (~CS_SM);
	set_Index_Address(CS_I8);
	outputb(p_des->reg[FS_And_Playback_Data_Format],p_des->direct_reg[Index_Data].indreg);
	//single cycle for format and clock frequency 8khz
	natb tmp;
	switch (format)
	{
		case U8	:	tmp = (khz == 8000)? 0x00 : 0x0c;
					break;
		case L16: 	tmp = (khz == 8000)? 0x40 : 0x4c;
					break;
	}
	p_des->reg[FS_And_Playback_Data_Format] |= tmp;
	set_Index_Address(CS_I8);
	outputb(p_des->reg[FS_And_Playback_Data_Format],p_des->direct_reg[Index_Data].indreg);
	//single dma channel
	p_des->reg[Interface_Configuration] |= CS_SDC;
	set_Index_Address(CS_I9);
	outputb(p_des->reg[Interface_Configuration],p_des->direct_reg[Index_Data].indreg);
	//setto calibrazione
	p_des->reg[Interface_Configuration] = ((p_des->reg[Interface_Configuration] & CS_CAL) | 0x18);
	set_Index_Address(CS_I9);
	outputb(p_des->reg[Interface_Configuration],p_des->direct_reg[Index_Data].indreg);
	reset_mce();
	reset_pmce();
	
	//aspetto calibrazione
	natb status;
	do{
		set_Index_Address(CS_I11);
		inputb(p_des->direct_reg[Index_Data].indreg,status);	
	}while (status != 0x00);
	//gestisco output
	natb volume;
	if (mod == MONO)
	{	
		p_des->reg[Left_DAC_Output_Control] &= (~CS_LDM); 
		p_des->reg[Right_DAC_Output_Control] &= (~CS_RDM);
		set_Index_Address(CS_I6);
		outputb(p_des->reg[Left_DAC_Output_Control],p_des->direct_reg[Index_Data].indreg);
		set_Index_Address(CS_I7);
		outputb(p_des->reg[Right_DAC_Output_Control],p_des->direct_reg[Index_Data].indreg);
		volume = 0x00;
		p_des->reg[Mono_Input_And_Output_Control] |= CS_MOM; 
		p_des->reg[Mono_Input_And_Output_Control] = ((p_des->reg[Mono_Input_And_Output_Control] & CS_MIA) | volume) ; 
		set_Index_Address(CS_I26);
		outputb(p_des->reg[Mono_Input_And_Output_Control],p_des->direct_reg[Index_Data].indreg);
	}else
	{
		p_des->reg[Left_DAC_Output_Control] |= CS_LDM; 
		p_des->reg[Right_DAC_Output_Control] |= CS_RDM;
		volume = 0x3f;
		p_des->reg[Right_DAC_Output_Control] = ((p_des->reg[Right_DAC_Output_Control] & R_DAC_A) | volume);
		p_des->reg[Left_DAC_Output_Control] = ((p_des->reg[Left_DAC_Output_Control] & L_DAC_A) | volume);
		set_Index_Address(CS_I6);
		outputb(p_des->reg[Left_DAC_Output_Control],p_des->direct_reg[Index_Data].indreg);
		set_Index_Address(CS_I7);
		outputb(p_des->reg[Right_DAC_Output_Control],p_des->direct_reg[Index_Data].indreg);
		p_des->reg[Mono_Input_And_Output_Control] &= (~CS_MOM); 
		set_Index_Address(CS_I26);
		outputb(p_des->reg[Mono_Input_And_Output_Control],p_des->direct_reg[Index_Data].indreg);		
	}
	p_des->set = true;
	sem_signal(p_des->CS_mutex);	
	return true;		
}

void stop_cs()
{
	suono* p_suono;
	p_suono = &cs_suono;
	p_suono->buffer = 0;
	p_suono->quanti = 0;
	sem_signal(p_suono->sync);
}


void start_cs()
{	
	c_writeconsole("sono nella play\n");
	suono* p_suono;
	p_suono = &cs_suono;
	dmacCH3* p_dmac;
	p_dmac = &dmac;
	CScard* p_des;
	p_des = &cs_card;

	natb upper_count;
	natb lower_count;
	addr fisico = trasforma(p_suono->buffer);
	natw rim = 4096 - ((natq)fisico & 0xfff);
	rim = (rim > p_suono->quanti)? p_suono->quanti : rim;
	natw baseRegister;
	natb tmp = (p_des->reg[FS_And_Playback_Data_Format] & 0xf0);
	switch (tmp)
	{
		case 0x00 : baseRegister = rim-1;
					break;
		case 0x40 : baseRegister = (rim/2)-1;
					break;
		case 0x10 :	baseRegister = (rim/2)-1;
					break;
		case 0x50 : baseRegister = (rim/4)-1;
					break;
	}
	lower_count = (natb)baseRegister;
	upper_count = (natb)(baseRegister >> 8);
	sem_wait(p_suono->mutex);
	//avvio primo trasferimento	
	sem_wait(p_dmac->mutex);
	//flog(LOG_INFO,"indirizzo virtuale %lx, fisico: %lx",p_suono->buffer,fisico);
	//flog(LOG_INFO,"quanti :%d,rim:%d,baseRegister:%x, upper count : %x, lower count : %x",p_suono->quanti,rim,baseRegister,upper_count,lower_count);
	//carico il numero di cacmpioni nei reigisti della cs
	set_lowerbase(lower_count);
	set_upperbase(upper_count);
	lower_count = (natb)(rim-1);
	upper_count = (natb)((rim-1) >> 8);
	set_dmac(fisico,lower_count,upper_count); //dico al controllore dma da quale ind e quanti BYTE deve passare
	p_suono->buffer += rim ;
	p_suono->quanti -= rim ;
	set_pen(1);
	c_writeconsole("Inizio riproduzione");
	sem_wait(p_suono->sync);
	set_pen(0);
	c_writeconsole("Fine riproduzione");
	mem_free(sysBuff);
	sem_signal(p_dmac->mutex);
	sem_signal(p_suono->mutex);
}

extern "C" bool c_command_cs(int cmd)
{
	CScard* p_des;
	p_des = &cs_card;
	if(!p_des->set)
	{
		c_writeconsole("settare scheda audio");
		return false;
	}
	switch (cmd){
		case START:
			start_cs();
			break;
		case PAUSE:
			p_des->direct_reg[Index_Address].valore |= CS_TRD;
			outputb(p_des->direct_reg[Index_Address].valore,p_des->direct_reg[Index_Address].indreg);
			set_pen(0);
			break;
		case PLAY:
			p_des->direct_reg[Index_Address].valore &= (~CS_TRD);
			outputb(p_des->direct_reg[Index_Address].valore,p_des->direct_reg[Index_Address].indreg);
			set_pen(1);
			break;
		case STOP:	
			stop_cs();
			break;
		default:
			c_writeconsole("Selezionare un comando valido \n ");
			return false;
			break;			
	}
	return true;
}

void estern_cs(int h)
{
	CScard* p_des;
	p_des = &cs_card;
	suono* p_suono;
	p_suono = &cs_suono;
	natb out = 0x00;
	for(;;)
	{
		//rispondo alla int
		outputb(out,p_des->direct_reg[Status].indreg);	
		if(p_suono->quanti >0)
		{
			natb upper_count;
			natb lower_count;
			addr fisico=trasforma(p_suono->buffer);
			natq rim = p_suono->quanti;
			rim = (rim > 4096)? 4096 : rim;
			natw baseRegister;
			natb tmp = (p_des->reg[FS_And_Playback_Data_Format] & 0xf0);
			switch (tmp)
			{
				case 0x00 : baseRegister = rim-1;
							break;
				case 0x40 : baseRegister = (rim/2)-1;
							break;
				case 0x10 :	baseRegister = (rim/2)-1;
							break;
				case 0x50 : baseRegister = (rim/4)-1;
							break;
			}
			lower_count = (natb)baseRegister;
			upper_count = (natb)(baseRegister >> 8);
			//flog(LOG_INFO,"sono nel pe,quanti:%d, rim: %d,id fisico:%x",p_suono->quanti,rim,fisico);	
			set_upperbase(upper_count);
			set_lowerbase(lower_count);
			lower_count = (natb)(rim-1);
			upper_count = (natb)((rim-1) >> 8);
			set_dmac(fisico,lower_count,upper_count);
			p_suono->buffer += rim ;
			p_suono->quanti -= rim;
		}else{
			sem_signal(p_suono->sync);
		}
		wfi();
	}
}

bool CS_init()
{
	CScard* p_des;
	p_des = &cs_card;

	suono* p_suono;
	p_suono = &cs_suono;
	natw base = 0x534; //da specifiche qemu
	p_des->presente = true;

	if(!CS_found(base))
	{
		flog(LOG_ERR,"cs4231a non trovata");
		return p_des->presente = false;
	}
	//inizializzo i 4 registri diretti
	for (int i = 0; i < 4; ++i){
		p_des->direct_reg[i].indreg = (ioaddr)(base+i);
	}
	// entro in mod 2
	p_des->reg[MODE_And_ID] |= CS_MODE2; 
	set_Index_Address(CS_I12);
	outputb(p_des->reg[MODE_And_ID],p_des->direct_reg[Index_Data].indreg);
	
	//salvo i valori al reset di alcuni registri
	p_des->reg[Left_ADC_Input_Control] = 		0x00;
	p_des->reg[Right_ADC_Input_Control] = 		0x00;
	p_des->reg[Left_AUX1_Input_Control] = 		0x88;
	p_des->reg[Right_AUX1_Input_Control] = 		0x88;
	p_des->reg[Left_AUX2_Input_Control] = 		0x88;
	p_des->reg[Right_AUX2_Input_Control] = 		0x88;
	p_des->reg[Left_DAC_Output_Control] = 		0x80;
	p_des->reg[Right_DAC_Output_Control] = 		0x80;
	p_des->reg[FS_And_Playback_Data_Format] = 	0x00;
	p_des->reg[Interface_Configuration] = 		0x08;
	p_des->reg[Pin_Control] = 					0x00;
	p_des->reg[Loopback_Control] = 				0x00;
	p_des->reg[Alternate_Feature_Enable_I] = 	0x00;
	p_des->reg[Alternate_Feature_Enable_II] = 	0x00;
	p_des->reg[Left_Line_Input_Control] = 		0x88;
	p_des->reg[Right_Line_Input_Control] = 		0x88;
	p_des->reg[Timer_Low_Base] = 				0x00;
	p_des->reg[Timer_High_Base] = 				0x00;
	p_des->reg[Alternate_Feature_Enable_III] = 	0x00;
	p_des->reg[Alternate_Feature_Status] = 		0x00;
	p_des->reg[Version_Chip_ID] = 				0xa0;
	p_des->reg[Mono_Input_And_Output_Control] = 0xa0;
	p_des->reg[Capture_Data_Format] = 			0x00;

	p_des->irq = 9; // piedino int da specifiche qemu
	p_des->CS_mutex = sem_ini(1);
	if (p_des->CS_mutex == 0xFFFFFFFF)
	{
		flog(LOG_ERR,"impossibile allocare mutex per cs4231a");
		return p_des->presente = false;
	}
	natl id;
	id = activate_pe(estern_cs, 0 , PRIO, LIV , p_des->irq);
	if (id == 0xFFFFFFFF)	
	{
		flog(LOG_ERR,"impossibile attivare pe per cs4231a");
		return p_des->presente = false;		
	}

	flog(LOG_INFO,"Scheda audio cs4231a inizializzata correttamente");
	
	p_suono->sync = sem_ini(0);
	if (p_suono->sync == 0xFFFFFFFF)
	{
		flog(LOG_ERR,"impossibile allocare sync per  suono di cs4231a");
		return p_des->presente = false;
	}	

	p_suono->mutex = sem_ini(1);
	if (p_suono->mutex == 0xFFFFFFFF)
	{
		flog(LOG_ERR,"impossibile allocare mutex per  suono di cs4231a");
		return p_des->presente = false;
	}

	return p_des->presente;
}

bool dmac_init()
{
	dmacCH3* p;
	p = &dmac;
	p->dmac_reg[Start_Address_Register] = 0x06;
	p->dmac_reg[Count_Register] = 0x07;
	p->dmac_reg[Single_Channel_Mask_Register] = 0x0a;
	p->dmac_reg[Mode_Register] = 0x0b;
	p->dmac_reg[FlipFlop_Reset_Register] = 0x0c;
	p->dmac_reg[Page_Register] = 0x82;

	p->mutex = sem_ini(1);
	if (p->mutex == 0xFFFFFFFF)
	{
		flog(LOG_ERR,"impossibile allocare mutex per dmac");
		return false;
	}
	return true;
}


////////////////////////////////////////////////////////////////////////////////
//                 INIZIALIZZAZIONE DEL SOTTOSISTEMA DI I/O                   //
////////////////////////////////////////////////////////////////////////////////

// inizializza i gate usati per le chiamate di IO
//
extern "C" void fill_io_gates(void);

extern "C" natl end;
// eseguita in fase di inizializzazione
//
extern "C" void cmain(int sem_io)
{

	fill_io_gates();
	mem_mutex = sem_ini(1);
	if (mem_mutex == 0xFFFFFFFF) {
		flog(LOG_ERR, "impossible creare semaforo mem_mutex");
		abort_p();
	}
	unsigned long long end_ = (unsigned long long)&end;
	end_ = (end_ + DIM_PAGINA - 1) & ~(DIM_PAGINA - 1);
	heap_init((void *)end_, DIM_IO_HEAP);
	if (!console_init())
		abort_p();
	if (!com_init())
		abort_p();
	if (!hd_init())
		abort_p();
	if (!dmac_init())
		flog(LOG_ERR,"dmac non inizializzato");
	if (!CS_init())
		flog(LOG_ERR,"Inizializzazione scheda audio cs4231a non riuscita");
	sem_signal(sem_io);
	terminate_p();
}
